#include "TMDlib_InterpolationKS.h"

//------------------------------------------------------------------------------
//
//   2D INTERPOLATION
//
//------------------------------------------------------------------------------
Interpolation2D::Interpolation2D(const vector<double> &x1vv, 
                             const vector<double> &x2vv, 
			     boost::multi_array<double, 2> ym,
                             gsl_interp_type int_type)
                             :_int_type(int_type),
                               n1(x1vv.size()), n2(x2vv.size()), y(ym),
			       x1v(x1vv), x2v(x2vv) {
    
  //GSL 2d interpolation:
  x1acc = gsl_interp_accel_alloc();
  x2acc = gsl_interp_accel_alloc();
  double gsltab[n1*n2];  

  _int2d_type = *gsl_interp2d_bilinear;
  spline2d = gsl_spline2d_alloc( &_int2d_type, n1, n2 );
  //x1 on x axis, x2 on y axis; set all x2 for one x1 then move to next x1
  for( unsigned int i=0; i<n1; i++ ){
   for( unsigned int j=0; j<n2; j++ ){
   double val = y[i][j];
    gsl_spline2d_set( spline2d, gsltab, i, j, val );
   }
  }

  gsl_spline2d_init( spline2d, &x1v[0], &x2v[0], gsltab, n1, n2 );
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
Interpolation2D::~Interpolation2D() {
    // clean the memory corresponding to pointers from spline and acc matrices
    gsl_interp_accel_free( x1acc );
    gsl_interp_accel_free( x2acc );
    gsl_spline2d_free( spline2d );
}

//-----------------------------------------------------------------------
vector<pair <double, double> > Interpolation2D::grid_limits() const {

  pair <double, double> x1_limits(*(x1v.begin()),*(x1v.rbegin()));
  pair <double, double> x2_limits(*(x2v.begin()),*(x2v.rbegin()));

  vector<pair <double, double> > limits;
  limits.push_back(x1_limits);
  limits.push_back(x2_limits);

  return limits;
}
  
//-----------------------------------------------------------------------
// return value from 2d spline interpolation
//-----------------------------------------------------------------------
double Interpolation2D::interp (const vector<double> &x) {

    double x1=x[0], x2=x[1];
    double val;
    bool first_x2out = true;
    _eval_status = 0;
   
   int gsl_status = gsl_spline2d_eval_e( spline2d, x1, x2, x1acc, x2acc, &val );
   if (gsl_status == GSL_EDOM) { _eval_status = _eval_status | 1; }
    return  val;
}
//------------------------------------------------------------------------------
//
//   3D INTERPOLATION
//
//------------------------------------------------------------------------------

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
CubicSpline3D::CubicSpline3D(const vector<double> &x1vv, 
                             const vector<double> &x2vv, 
                             const vector<double> &x3vv, 
			     boost::multi_array<double, 3> ym)
                             : n1(x1vv.size()), n2(x2vv.size()),n3(x3vv.size()),
			      y(ym), x1v(x1vv), x2v(x2vv), x3v(x3vv) {

   //Old Sebastian's code gsl 1d 
   // acc.resize(n1);
   // spline.resize(n1);

   // // interpolate in z for all points of (x1, x2) grid
   // for (unsigned i=0;i<n1;i++) {    // x1
   //   for (unsigned j=0;j<n2;j++) {  // x2
   //     acc[i].push_back(gsl_interp_accel_alloc ());
   //     spline[i].push_back(gsl_spline_alloc (gsl_interp_linear, n3));
   //     vector<double> y3;  
   //     for (unsigned k=0;k<n3;k++) 
   //       y3.push_back(y[i][j][k]); // x3
   //     gsl_spline_init (spline[i][j], &x3v[0], &y3[0], n3);
   //  }
   // }
   // ---

    //create n3 2-dimensional interpolations for each x3
    x1acc.resize(n3);
    x2acc.resize(n3);
    spline2d.resize(n3);
    vector<double> gsltabvec[n3]; //vector of tables needed for gsl 2d interpolation
    for(int i=0; i<n3; i++){gsltabvec[i].resize(n1*n2);}

   //create n3 gsl 2d objects
   for( unsigned int k=0; k<n3; k++ ){
    x1acc[k] = gsl_interp_accel_alloc();
    x2acc[k] = gsl_interp_accel_alloc();
    spline2d[k] = gsl_spline2d_alloc( gsl_interp2d_bilinear, n1, n2 );
    for( unsigned int i=0; i<n1; i++ ){
      for( unsigned int j=0; j<n2; j++ ){
       double val = y[i][j][k]; 
       gsl_spline2d_set( spline2d[k], &gsltabvec[k][0], i, j, val );
      }
    }
    gsl_spline2d_init( spline2d[k], &x1v[0], &x2v[0], &gsltabvec[k][0], n1, n2 );
   }

}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
CubicSpline3D::~CubicSpline3D() {
    
    // clean the memory corresponding to pointers from spline and acc matrices
    for( unsigned int k=0; k<n3; k++ ){
     gsl_interp_accel_free( x1acc[k] );
     gsl_interp_accel_free( x2acc[k] );
     gsl_spline2d_free( spline2d[k] );
    }
}

  
//-----------------------------------------------------------------------
/// return value from 3d interpolation
/// get result from interpolation in x1-x2 for each x3 and use it to do
/// interpolations in x3
//-----------------------------------------------------------------------
double CubicSpline3D::interp (const vector<double> &x) {

    double x1=x[0], x2=x[1], x3=x[2];
    double val;
    _eval_status = 0;

    //get values from n3 interpolations in x1-x2 plane
    double y3[n3];
    for( unsigned int k=0; k<n3; k++ ){
     int gsl_status = gsl_spline2d_eval_e( spline2d[k], x1, x2, x1acc[k], x2acc[k], &y3[k] );
     if( gsl_status == GSL_EDOM && !(_eval_status & 2) ) { _eval_status = _eval_status | 2; }
    }

    //create interpolation in x3
    gsl_interp_accel* acc1d = gsl_interp_accel_alloc();
    gsl_spline* spline1d = gsl_spline_alloc( gsl_interp_linear, n3 );
    gsl_spline_init( spline1d, &x3v[0], y3, n3 );

    int gsl_status = gsl_spline_eval_e( spline1d, x3, acc1d, &val );
    if( gsl_status == GSL_EDOM ) { _eval_status = _eval_status | 1; } 
   
    gsl_interp_accel_free( acc1d );
    gsl_spline_free( spline1d );

    //cout << "x=" << x1 << " kt=" << x2 << " mu=" << x3 << " xg= " << val << endl;
    return  val;
}
//-----------------------------------------------------------------------
vector<pair <double, double> > CubicSpline3D::grid_limits() const {


  pair <double, double> x1_limits(*(x1v.begin()),*(x1v.rbegin()));
  pair <double, double> x2_limits(*(x2v.begin()),*(x2v.rbegin()));
  pair <double, double> x3_limits(*(x3v.begin()),*(x3v.rbegin()));

  vector<pair <double, double> > limits;
  limits.push_back(x1_limits);
  limits.push_back(x2_limits);
  limits.push_back(x3_limits);

  return limits;
}
